/*
  ==============================================================================

   This file is part of the Plugin Consultant - VST API Logger
   Copyright 2005 by Ofer Shem Tov.

  ------------------------------------------------------------------------------

   Plugin Consultant can be redistributed and/or modified under the terms of the
   GNU General Public License, as published by the Free Software Foundation;
   either version 2 of the License, or (at your option) any later version.

   Plugin Consultant is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Plugin Consultant; if not, visit www.gnu.org/licenses or write to the
   Free Software Foundation, Inc., 59 Temple Place, Suite 330,
   Boston, MA 02111-1307 USA

  ==============================================================================
*/

#ifdef _MSC_VER
#pragma warning (disable : 4996)
#endif

#include <aeffectx.h>
#ifdef VST_2_4_EXTENSIONS
  #include <AEffEditor.h>
#else
  #include <AEffEditor.hpp>
#endif
#include <juce.h>

// HS: crude distinction between the JUCE versions I use, which are
// V1.21 (doesn't have Ogg Vorbis support) and V1.42 (does have it,
// along with gazillions of other changes that unfortunately are totally
// incompatible with the earlier version <sigh>...).
#ifndef IS_NEW_JUCE
  #ifdef JUCE_USE_OGGVORBIS
    #define IS_NEW_JUCE 1
  #else
    #define IS_NEW_JUCE 0
  #endif
#endif

#include "PCLogger.h"

#ifdef JUCE_WIN32
#undef T
#undef Rectangle
#define Array tempArray
#include <windows.h>
#undef Array
#define T(x) x 
#endif

enum {

	pcLogMax = 1024
};

//-----------------------------------------------------------------------------------------
// VST C functions
enum vstFunction {
	vstDispatcher,
	vstAudioMaster,
	vstProcess,
	vstProcessReplacing,
	vstSetParameter,
	vstGetParameter,
#ifdef VST_2_4_EXTENSIONS
    vstProcessDoubleReplacing,
#endif
	vstNumFunctions
};
//-----------------------------------------------------------------------------------------
static char* vstFunctionNames[] =
{
	T("dispatcher"),
	T("audioMaster"),
#ifdef VST_2_4_EXTENSIONS
	T("process (deprecated)"),
#else
	T("process"),
#endif
	T("processReplacing"),
	T("setParameter"),
	T("getParameter"),
#ifdef VST_2_4_EXTENSIONS
    T("processDoubleReplacing"),
#endif
};

//-----------------------------------------------------------------------------------------
// possible dispatcher ptr types
enum PtrType {

	ptrNone,
	ptrChar,
	ptrERect,
	ptrPtrERect,
	ptrPtrVoid,
	ptrVoid,
	ptrEvents,
	ptrPinProperties,
	ptrParameterProperties,
	ptrTimeInfo,
	ptrWindow,
	ptrFileSelect,
	ptrAudioFile,
	ptrOfflineTask,
	ptrVariableIo,
	ptrSpeakerArrangement,
	ptrMidiProgramName,
	ptrMidiProgramCategory,
	ptrMidiKeyName,
	ptrPatchChunkInfo,
};

//-----------------------------------------------------------------------------------------
// information about a VST opcode
struct opCodeDebugInfo {

	char*	opCodeName;	// opcode name
	char*	longName;	// long return value name or NULL if no return value
	char*	indexName;	// index parameter name or NULL if no index parameter
	char*	valueName;	// value parameter name or NULL if no value parameter
	bool	ptrReturn;	// true if ptr is a return value
	PtrType	ptrType;	// type of data ptr is pointing to, ptrNone if no ptr parameter
	char*	ptrName;	// name of ptr parameter or NULL if no ptr parameter
	char*	optName;	// float parameter name or NULL if no float parameter
};

//-----------------------------------------------------------------------------------------
static struct opCodeDebugInfo effOpCodes[] =
{
	// name							long		index		value		ptr-return	ptr-type		ptr			float
	{ T("Open"),					NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("Close"),					NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("SetProgram"),				NULL,		NULL,		T("program"),false,		ptrNone,		NULL,		NULL },
	{ T("GetProgram"),				T("program"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("SetProgramName"),			NULL,		NULL,		NULL,		false,		ptrChar,		T("name"),	NULL },
	{ T("GetProgramName"),			NULL,		NULL,		NULL,		true,		ptrChar,		T("name"),	NULL },
	{ T("GetParamLabel"),			NULL,		T("parameter"),NULL,	true,		ptrChar,		T("label"),	NULL },
	{ T("GetParamDisplay"),			NULL,		T("parameter"),NULL,	true,		ptrChar,		T("value"),	NULL },
	{ T("GetParamName"),			NULL,		T("parameter"),NULL,	true,		ptrChar,		T("name"),	NULL },
	{ T("GetVu (deprecated)"),		T("level"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("SetSampleRate"),			NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		T("sampleRate") },
	{ T("SetBlockSize"),			NULL,		NULL,		T("blockSize"),false,	ptrNone,		NULL,		NULL },
	{ T("MainsChanged"),			NULL,		NULL,		T("state"),	false,		ptrNone,		NULL,		NULL },
	{ T("EditGetRect"),				T("long"),	NULL,		NULL,		true,		ptrPtrERect,	"rect",		NULL },
	{ T("EditOpen"),				T("long"),	NULL,		NULL,		false,		ptrVoid,		T("window"),NULL },
	{ T("EditClose"),				NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("EditDraw (deprecated)"),	NULL,		NULL,		NULL,		false,		ptrERect,		T("rect"),	NULL },
	{ T("EditMouse (deprecated)"),	T("long"),	T("x"),		T("y"),		false,		ptrNone,		NULL,		NULL },
	{ T("EditKey (deprecated)"),	T("long"),	NULL,		T("keyCode"),false,		ptrNone,		NULL,		NULL },
	{ T("EditIdle"),				NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("EditTop (deprecated)"),	NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("EditSleep (deprecated)"),	NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("Identify (deprecated)"),	T("CCONST"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetChunk"),				T("dataSize"),T("isPreset"),	NULL,false,		ptrPtrVoid,		T("data"),	NULL },
	{ T("SetChunk"),				T("long"),	T("isPreset"),	T("byteSize"),false,ptrVoid,		T("data"),	NULL },
	{ T("ProcessEvents"),			T("long"),	NULL,		NULL,		false,		ptrEvents,		T("events"),NULL },
	{ T("CanBeAutomated"),			T("bool"),	T("parameter"),NULL,	false,		ptrNone,		NULL,		NULL },
	{ T("String2Parameter"),		T("bool"),	T("parameter"),NULL,	false,		ptrChar,		T("text"),	NULL },
	{ T("GetNumProgramCategories (dep.)"),	T("categories"),NULL,	NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetProgramNameIndexed"),	T("bool"),	T("index"),	T("category"),true,		ptrChar,		T("text"),	NULL },
	{ T("CopyProgram (deprecated)"), T("bool"),	T("destination"),NULL,	false,		ptrNone,		NULL,		NULL },
	{ T("ConnectInput (deprecated)"), NULL,		T("index"),	T("state"),	false,		ptrNone,		NULL,		NULL },
	{ T("ConnectOutput (deprecated)"), NULL,	T("index"),	T("state"),	false,		ptrNone,		NULL,		NULL },
	{ T("GetInputProperties"),		T("bool"),	T("index"),	NULL,		false,		ptrPinProperties,T("properties"),NULL },
	{ T("GetOutputProperties"),		T("bool"),	T("index"),	NULL,		false,		ptrPinProperties,T("properties"),NULL },
	{ T("GetPlugCategory"),			T("category"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetCurrentPosition (dep.)"),	T("long"),	NULL,	NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetDestinationBuffer (dep.)"),	T("float*"),NULL,	NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("OfflineNotify"),			T("bool"),	T("start"),	T("numAudioFiles"),false,ptrAudioFile,	T("audioFiles"),NULL },
	{ T("OfflinePrepare"),			T("bool"),	NULL,		T("count"),	false,		ptrOfflineTask,	T("offline"),NULL },
	{ T("OfflineRun"),				T("bool"),	NULL,		T("count"),	false,		ptrOfflineTask,	T("offline"),NULL },
	{ T("ProcessVarIo"),			T("bool"),	NULL,		NULL,		false,		ptrVariableIo,	T("varIo"),	NULL },
	{ T("SetSpeakerArrangement"),	T("bool"),	NULL,		T("pluginInput"),false,	ptrSpeakerArrangement,T("pluginOutput"),NULL },
	{ T("SetBlockSizeAndSampleRate (dep.)"),NULL,	NULL,	T("blockSize"),false,	ptrNone,		NULL,		T("sampleRate") },
	{ T("SetBypass"),				T("bool"),	NULL,		T("onOff"),	false,		ptrNone,		NULL,		NULL },
	{ T("GetEffectName"),			T("bool"),	NULL,		NULL,		true,		ptrChar,		T("name"),	NULL },
	{ T("GetErrorText (deprecated)"), T("bool"),	NULL,	NULL,		true,		ptrChar,		T("text"),	NULL },
	{ T("GetVendorString"),			T("bool"),	NULL,		NULL,		true,		ptrChar,		T("text"),	NULL },
	{ T("GetProductString"),		T("bool"),	NULL,		NULL,		true,		ptrChar,		T("text"),	NULL },
	{ T("GetVendorVersion"),		T("version"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("VendorSpecific"),			T("long"),	T("lArg1"),	T("lArg2"),	false,		ptrVoid,		T("prtArg"),T("floatArg") },
	{ T("CanDo"),					T("long"),	NULL,		NULL,		false,		ptrChar,		T("text"),	NULL },
	{ T("GetTailSize"),				T("long"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("Idle (deprecated)"),		T("long"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetIcon (deprecated)"),	T("void*"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("SetViewPosition (deprecated)"), T("bool"),	T("x"),	T("y"),		false,		ptrNone,		NULL,		NULL },
	{ T("GetParameterProperties"),	T("bool"),	T("index"),	NULL,		false,		ptrParameterProperties, T("properties"),NULL },
	{ T("KeysRequired (deprecated)"),	T("bool"),	NULL,	NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetVstVersion"),			T("version"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("EditKeyDown"),				T("long"),	T("character"),T("virt"),false,		ptrNone,		NULL,		T("modifier") },
	{ T("EditKeyUp"),				T("long"),	T("character"),T("virt"),false,		ptrNone,		NULL,		T("modifier") },
	{ T("SetEditKnobMode"),			T("long"),	NULL,		T("value"),	false,		ptrNone,		NULL,		NULL },
	{ T("GetMidiProgramName"),		T("programIndexes"),T("MidiChannel"),NULL,false,ptrMidiProgramName,T("program"),NULL },
	{ T("GetCurrentMidiProgram"),	T("programIndex"),	NULL,NULL,		false,		ptrMidiProgramName,T("program"),NULL },
	{ T("GetMidiProgramCategory"),	T("categoryIndexes"),NULL,NULL,		false,		ptrMidiProgramCategory,T("category"),NULL },
	{ T("HasMidiProgramsChanged"),	T("bool"),	T("MidiChannel"),NULL,	false,		ptrNone,		NULL,		NULL },
	{ T("GetMidiKeyName"),			T("bool"),	NULL,		NULL,		false,		ptrMidiKeyName,	T("key"),	NULL },
	{ T("BeginSetProgram"),			T("bool"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("EndSetProgram"),			T("bool"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetSpeakerArrangement"),	T("bool"),	NULL,		T("pluginInput"),false,	ptrSpeakerArrangement,T("pluginOutput"),NULL },
	{ T("ShellGetNextPlugin"),		T("uniqueID"),NULL,		NULL,		true,		ptrChar,		T("name"),	NULL },
	{ T("StartProcess"),			T("long"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("StopProcess"),				T("long"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("SetTotalSampleToProcess"),	T("long"),	NULL,		T("value"),	false,		ptrNone,		NULL,		NULL },
	{ T("SetPanLaw"),				T("bool"),	NULL,		T("type"),	false,		ptrNone,		NULL,		T("gain") },
	{ T("BeginLoadBank"),			T("long"),	NULL,		NULL,		false,		ptrPatchChunkInfo,T("patchChunkInfo"),NULL },
	{ T("BeginLoadProgram"),		T("long"),	NULL,		NULL,		false,		ptrPatchChunkInfo,T("patchChunkInfo"),NULL },
	{ T("SetProcessPrecision"),		T("long"),	T("Precision"),	NULL,	false,		ptrNone,		NULL,		NULL },
	{ T("GetNumMidiInputChannels"),	T("long"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetNumMidiOutputChannels"),T("long"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
};

//-----------------------------------------------------------------------------------------
static struct opCodeDebugInfo audioMasterOpCodes[] =
{
	// name								long		index		value		ptr-return	ptr-type		ptr			float
	{ T("Automate"),					NULL,		T("index"),	NULL,		false,		ptrNone,		NULL,		T("value") },
	{ T("Version"),						T("version"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("CurrentId"),					T("Unique ID"),NULL,	NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("Idle"),						NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("PinConnected (deprecated)"),	T("connected"),T("pin index"),T("isOutput"),	false,		ptrNone,		NULL,		NULL },
	{ T("ThereIsNoSuchOpCode"),			NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("WantMidi (deprecated)"),		NULL,		NULL,		T("filter"),false,		ptrNone,		NULL,		NULL },
	{ T("GetTime"),						T("TimeInfo"),NULL,		T("filter"),false,		ptrNone,		NULL,		NULL },
	{ T("ProcessEvents"),				T("bool"),	NULL,		NULL,		false,		ptrEvents,		T("events"),NULL },
	{ T("SetTime (deprecated)"),		NULL,		NULL,		T("filter"),false,		ptrTimeInfo,	T("TimeInfo"),NULL },
	{ T("TempoAt (deprecated)"),		T("tempo"),	NULL,		T("pos"),	false,		ptrNone,		NULL,		NULL },
	{ T("GetNumAutomatableParameters (dep.)"), T("long"), NULL,	NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetParameterQuantization (dep."), T("long"), NULL,		T("index"),	false,		ptrNone,		NULL,		NULL },
	{ T("IOChanged"),					T("bool"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("NeedIdle (deprecated)"),		T("bool"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("SizeWindow"),					T("bool"),	T("width"),	T("height"),false,		ptrNone,		NULL,		NULL },
	{ T("GetSampleRate"),				T("sampleRate"),NULL,	NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetBlockSize"),				T("blockSize"),NULL,	NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetInputLatency"),				T("latency"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetOutputLatency"),			T("latency"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetPreviousPlug (deprecated)"), T("effect"),NULL,		T("inpuPin"),false,		ptrNone,		NULL,		NULL },
	{ T("GetNextPlug (deprecated)"),	T("effect"),NULL,		T("outputPin"),false,	ptrNone,		NULL,		NULL },
	{ T("WillReplaceOrAccumulate (dep.)"), T("long"), NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetCurrentProcessLevel"),		T("long"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetAutomationState"),			T("long"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("OfflineStart"),				T("bool"),	T("numNewAudioFiles"),T("numAudioFiles"),false,ptrAudioFile,T("audioFiles"),NULL },
	{ T("OfflineRead"),					T("bool"),	T("readSource"),T("option"),false,	ptrOfflineTask,	T("offline"),NULL },
	{ T("OfflineWrite"),				T("bool"),	NULL,		T("option"),false,		ptrOfflineTask,	T("offline"),NULL },
	{ T("OfflineGetCurrentPass"),		T("long"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("OfflineGetCurrentMetaPass"),	T("long"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("SetOutputSampleRate (dep.)"),	NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		T("sampleRate") },
	{ T("GetSpeakerArrangement (dep.)"), NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("GetVendorString"),				T("bool"),	NULL,		NULL,		true,		ptrChar,		T("bool"),	NULL },
	{ T("GetProductString"),			T("bool"),	NULL,		NULL,		true,		ptrChar,		T("bool"),	NULL },
	{ T("GetVendorVersion"),			T("version"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("VendorSpecific"),				T("long"),	T("lArg1"),	T("lArg2"),	false,		ptrVoid,		T("prtArg"),T("floatArg") },
	{ T("SetIcon (deprecated)"),		NULL,		NULL,		NULL,		false,		ptrVoid,		T("icon"),	NULL },
	{ T("CanDo"),						T("long"),	NULL,		NULL,		false,		ptrChar,		T("bool"),	NULL },
	{ T("GetLanguage"),					T("language"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("OpenWindow (deprecated)"),		T("bool"),	NULL,		NULL,		false,		ptrWindow,		T("window"),NULL },
	{ T("CloseWindow (deprecated)"),	T("bool"),	NULL,		NULL,		false,		ptrWindow,		T("window"),NULL },
	{ T("GetDirectory"),				T("plugDir"),NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("UpdateDisplay"),				T("bool"),	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("BeginEdit"),					T("bool"),	T("index"),	NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("EndEdit"),						T("bool"),	T("index"),	NULL,		false,		ptrNone,		NULL,		NULL },
	{ T("OpenFileSelector"),			T("bool"),	NULL,		NULL,		false,		ptrFileSelect,	T("VstFileSelect*"),NULL },
	{ T("CloseFileSelector"),			T("bool"),	NULL,		NULL,		false,		ptrFileSelect,	T("VstFileSelect*"),NULL },
	{ T("EditFile (deprecated)"),		NULL,		NULL,		NULL,		false,		ptrChar,		T("xml"),	NULL },
	{ T("GetChunkFile (deprecated)"),	T("bool"),	NULL,		NULL,		false,		ptrVoid,		T("nativePath"),NULL },
	{ T("GetInputSpeakerArrangement (dep.)"), T("speakerArrangement"),NULL,NULL,false,	ptrNone,		NULL,		NULL },
};

//-----------------------------------------------------------------------------------------

struct DynamicDebugInfo {

	long count;		// how many time the opcode/function was called
	bool enabled;	// log is enabled for this opcode/function
};

enum {
#if VST_2_4_EXTENSIONS
	numEffOpCodes = effGetNumMidiOutputChannels + 1,
	numAudioMasterOpCodes = DECLARE_VST_DEPRECATED (audioMasterGetInputSpeakerArrangement) + 1,	// audioMaster op codes
#else
	numEffOpCodes = effNumV2_3Opcodes,									// eff op codes
	numAudioMasterOpCodes = audioMasterGetInputSpeakerArrangement + 1,	// audioMaster op codes
#endif
};

// arrays of all dynamic debug info for all VST functions and opCodes
static struct DynamicDebugInfo vstFunctionsLogInfo[vstNumFunctions];
static struct DynamicDebugInfo vstDispatcherLogInfo[numEffOpCodes ];
static struct DynamicDebugInfo vstAudioMasterLogInfo[numAudioMasterOpCodes];

//-----------------------------------------------------------------------------------------

struct vstOpCodeGroupInfo {

	int numOpCodes;					// number of op codes in group
	char* name;						// opCodes group name
	DynamicDebugInfo* debugInfo;	// group debug info
};

// Groups of OpCodes supported by the logger
static vstOpCodeGroupInfo OpCodeGroupsInfo[PCLogger::vstNumOpCodeGroups] = {

	{ vstNumFunctions, T("vstAEffectFunctions"), vstFunctionsLogInfo },
	{ numEffOpCodes , T("vstDispatcherOpCodes"), vstDispatcherLogInfo },
	{ numAudioMasterOpCodes, T("vstAudioMasterOpCodes"), vstAudioMasterLogInfo }
};

//-----------------------------------------------------------------------------------------
void initLogInfo( DynamicDebugInfo* info, int numInfos )
{
	for ( int i = 0; i < numInfos; i++ )
	{
		info->count = 0;
		info->enabled = true;
		info++;
	};
}

void initLogCounters( DynamicDebugInfo* info, int numInfos )
{
	for ( int i = 0; i < numInfos; i++ )
	{
		info->count = 0;
		info++;
	};
}

inline DynamicDebugInfo* getLogInfo( PCLogger::vstOpCodeGroup group, long opCode )
{
	if ( group < 0 || group >= PCLogger::vstNumOpCodeGroups )
		return NULL;

	return &OpCodeGroupsInfo[group].debugInfo[opCode];
}

//-----------------------------------------------------------------------------------------
void PCLogger::logMessage ( const String& message )
{
	log( message );
}

//-----------------------------------------------------------------------------------------
void PCLogger::log( const char* message )
{
	char buf[pcLogMax];
	int nextChar = 0;
	
	if ( isLogTime() )
	{
		Time time = Time::getCurrentTime ();
		nextChar = sprintf ( buf, T("%02d:%02d:%02d.%03d " ), time.getHours(), time.getMinutes(), time.getSeconds(), time.getMilliseconds() );
	}
	sprintf( &buf[nextChar], T("(%X) %s:%s"), Thread::getCurrentThreadId(), (const char*)plugName, message );
#if WIN32
	outputDebugString ( buf );
#endif
#if MAC
    Logger::outputDebugString( buf);
#endif
}

//-----------------------------------------------------------------------------------------
void PCLogger::init( const String& name )
{
	logOn = true;

	logParams = true;
	logRetParams = true;
	logTime = false;

	plugName = name;

	// init dynamic debug info data
	initLogInfo( vstFunctionsLogInfo, vstNumFunctions );
	initLogInfo( vstDispatcherLogInfo, numEffOpCodes );
	initLogInfo( vstAudioMasterLogInfo, numAudioMasterOpCodes );

	Logger::setCurrentLogger( this );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstDispatcher(long opCode, long index, long value, void *ptr, float opt)
{
	// if opCode is newer than VST SDK 2.3 then ignore
	if ( opCode < effOpen || opCode >= numEffOpCodes )
		return;

	// increment counter for dispatcher and opCode
	vstFunctionsLogInfo[vstDispatcher].count++;
	vstDispatcherLogInfo[opCode].count++;

	// log only if function and opCode logging are enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstDispatcher].enabled || !vstDispatcherLogInfo[opCode].enabled )
		return;

	char buf[pcLogMax];
	sprintf( buf, T("eff%s "), effOpCodes[opCode].opCodeName );

	// log parameters if enabled
	if ( isLogParameters() )
	{
		int nextChar = strlen( buf );

		// get parameters name and values if relevant for this opCode
		if ( effOpCodes[opCode].indexName != NULL )
				nextChar += sprintf( &buf[nextChar], T("%s=%d "), effOpCodes[opCode].indexName, (int) index );

		if ( effOpCodes[opCode].valueName != NULL )
			nextChar += sprintf( &buf[nextChar], T("%s=%d "), effOpCodes[opCode].valueName, (int) value );

		// even if there is a ptr value don't log it if its a return value
		if ( effOpCodes[opCode].ptrName != NULL && !effOpCodes[opCode].ptrReturn )
		{
			if ( effOpCodes[opCode].ptrType == ptrChar )
				nextChar += sprintf( &buf[nextChar], T("%s=%s "), effOpCodes[opCode].ptrName, (char*) ptr );
			else
				nextChar += sprintf( &buf[nextChar], T("%s=%p "), effOpCodes[opCode].ptrName, ptr );
		}

		if ( effOpCodes[opCode].optName != NULL )
			nextChar += sprintf( &buf[nextChar], T("%s=%f "), effOpCodes[opCode].optName, opt );
	}

	log( buf );
}

//-----------------------------------------------------------------------------------------
void	PCLogger::logVstDispatcherReturn( long opCode, void *ptr, long result )
{
	// do nothing if not logging return parameters
	if ( !isLogReturnParameters() )
		return;

	// if opCode is newer than VST SDK 2.3 then ignore
	if ( opCode < effOpen || opCode >= numEffOpCodes )
		return;

	// log only if function and opCode logging are enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstDispatcher].enabled || !vstDispatcherLogInfo[opCode].enabled )
		return;

	// do nothing if this opCode don't have any return values
	if ( effOpCodes[opCode].longName == NULL && !effOpCodes[opCode].ptrReturn )
		return;

	char buf[pcLogMax];
	sprintf( buf, T("eff%s returned "), effOpCodes[opCode].opCodeName );

	int nextChar = strlen( buf );

	// see if a long result is relevant for this opCode
	if ( effOpCodes[opCode].longName != NULL )
		nextChar += sprintf( &buf[nextChar], T("%s=%d "), effOpCodes[opCode].longName, (int) result );

	// log only if its a ptr return value and a name for the pointer paramter
	if ( effOpCodes[opCode].ptrName != NULL && effOpCodes[opCode].ptrReturn )
	{
		if ( effOpCodes[opCode].ptrType == ptrChar )
			nextChar += sprintf( &buf[nextChar], T("%s=%s"), effOpCodes[opCode].ptrName, (char*) ptr );
		else if ( effOpCodes[opCode].ptrType == ptrPtrERect && ptr != NULL && (*(ERect**)ptr) != NULL )
			nextChar += sprintf( &buf[nextChar], T("%s=top(%d), left(%d), bottom(%d), right(%d) "),
													effOpCodes[opCode].ptrName,
													(*(ERect**)ptr)->top, (*(ERect**)ptr)->left,
													(*(ERect**)ptr)->bottom, (*(ERect**)ptr)->right);
		else
			nextChar += sprintf( &buf[nextChar], T("%s=%p"), effOpCodes[opCode].ptrName, ptr );
	}

	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstAudioMaster( long opCode, long index, long value, void *ptr, float opt )
{
	// if opCode is newer than VST SDK 2.3 then ignore it
	if ( opCode < audioMasterAutomate  || opCode >= numAudioMasterOpCodes )
		return;

	// increment counter for audioMaster and opCode
	vstFunctionsLogInfo[vstAudioMaster].count++;
	vstAudioMasterLogInfo[opCode].count++;

	// log only if function and opCode logging are enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstAudioMaster].enabled || !vstAudioMasterLogInfo[opCode].enabled )
		return;

	char buf[pcLogMax];
	sprintf( buf, T("audioMaster%s "), audioMasterOpCodes[opCode].opCodeName );

	// log parameters if enabled
	if ( isLogParameters() )
	{
		int nextChar = strlen( buf );

		// get parameters name and values if relevant for this opCode
		if ( audioMasterOpCodes[opCode].indexName != NULL )
				nextChar += sprintf( &buf[nextChar], T("%s=%d "), audioMasterOpCodes[opCode].indexName, (int) index );

		if ( audioMasterOpCodes[opCode].valueName != NULL )
			nextChar += sprintf( &buf[nextChar], T("%s=%d "), audioMasterOpCodes[opCode].valueName, (int) value );

		// even if there is a ptr value don't log it if its a return value
		if ( audioMasterOpCodes[opCode].ptrName != NULL && !audioMasterOpCodes[opCode].ptrReturn )
		{
			if ( audioMasterOpCodes[opCode].ptrType == ptrChar )
				nextChar += sprintf( &buf[nextChar], T("%s=%s "), audioMasterOpCodes[opCode].ptrName, (char*) ptr );
			else
				nextChar += sprintf( &buf[nextChar], T("%s=%p "), audioMasterOpCodes[opCode].ptrName, ptr );
		}

		if ( audioMasterOpCodes[opCode].optName != NULL )
			nextChar += sprintf( &buf[nextChar], T("%s=%f "), audioMasterOpCodes[opCode].optName, opt );
	}

	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstAudioMasterReturn( long opCode, void *ptr, long result )
{
	// do nothing if not logging return parameters
	if ( !isLogReturnParameters() )
		return;

	// if opCode is newer than VST SDK 2.3 then ignore it
	if ( opCode < audioMasterAutomate  || opCode >= numAudioMasterOpCodes )
		return;

	// log only if function and opCode logging are enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstAudioMaster].enabled || !vstAudioMasterLogInfo[opCode].enabled )
		return;

	// do nothing if this opCode don't have any return values
	if ( audioMasterOpCodes[opCode].longName == NULL && !audioMasterOpCodes[opCode].ptrReturn )
		return;

	char buf[pcLogMax];
	sprintf( buf, T("audioMaster%s returned "), audioMasterOpCodes[opCode].opCodeName );

	int nextChar = strlen( buf );

	// see if a long result is relevant for this opCode
	if ( audioMasterOpCodes[opCode].longName != NULL )
		nextChar += sprintf( &buf[nextChar], T("%s=%d "), audioMasterOpCodes[opCode].longName, (int) result );

	// log only if its a ptr return value and a name for the pointer paramter
	if ( audioMasterOpCodes[opCode].ptrName != NULL && audioMasterOpCodes[opCode].ptrReturn )
	{
		if ( audioMasterOpCodes[opCode].ptrType == ptrChar )
			nextChar += sprintf( &buf[nextChar], T("%s=%s"), audioMasterOpCodes[opCode].ptrName, (char*) ptr );
		else
			nextChar += sprintf( &buf[nextChar], T("%s=%p"), audioMasterOpCodes[opCode].ptrName, ptr );
	}

	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstProcess( float **inputs, float **outputs, long sampleframes, bool replacing )
{
	int func;

	if ( replacing )
		func = vstProcessReplacing;
	else
		func = vstProcess;

	// increment counter for process/ProcessReplacing
	vstFunctionsLogInfo[func].count++;
	
	// log only if function is enabled
	if ( !isOn() || !vstFunctionsLogInfo[func].enabled )
		return;

	char buf[pcLogMax];

	if ( replacing )
		strcpy( buf, T("processReplacing") );
	else
		strcpy( buf, T("process") );

	// log parameters if enabled
	if ( isLogParameters() )
		sprintf( &buf[strlen( buf )], T("(inputs:%p, outputs:%p, sampleFrames:%d)"), inputs, outputs, (int) sampleframes);
	
	log( buf );
}

//-----------------------------------------------------------------------------------------
#ifdef VST_2_4_EXTENSIONS
void PCLogger::logVstProcessDoubleReplacing( double **inputs, double **outputs, VstInt32 sampleframes )
{
	int func = vstProcessDoubleReplacing;

	// increment counter for processDoubleReplacing
	vstFunctionsLogInfo[func].count++;
	
	// log only if function is enabled
	if ( !isOn() || !vstFunctionsLogInfo[func].enabled )
		return;

	char buf[pcLogMax];

	strcpy( buf, T("processDoubleReplacing") );

	// log parameters if enabled
	if ( isLogParameters() )
		sprintf( &buf[strlen( buf )], T("(inputs:%p, outputs:%p, sampleFrames:%d)"), inputs, outputs, (int) sampleframes);
	
	log( buf );
}
#endif

//-----------------------------------------------------------------------------------------
void PCLogger::logVstSetParameter( long index, float parameter )
{
	// increment counter for set parameter
	vstFunctionsLogInfo[vstSetParameter].count++;

	// log only if function is enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstSetParameter].enabled )
		return;

	char buf[pcLogMax];
	strcpy( buf, T("setParameter") );

	// log parameters if enabled
	if ( isLogParameters() )
		sprintf( &buf[strlen( buf )], T("(index: %d, parameter: %f)"), (int) index, parameter );

	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstGetParameter( long index )
{
	// increment counter for get parameter
	vstFunctionsLogInfo[vstGetParameter].count++;

	// log only if function is enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstGetParameter].enabled )
		return;

	char buf[pcLogMax];
	strcpy( buf, T("getParameter") );

	// log parameters if enabled
	if ( isLogParameters() )
		sprintf( &buf[strlen( buf )], T("(index: %d)"), (int) index );

	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstGetParameterReturn( float parameter )
{
	// do nothing if not logging return parameters
	if ( !isLogReturnParameters() )
		return;

	// log only if function is enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstGetParameter].enabled )
		return;

	char buf[pcLogMax];
	sprintf( buf, T("getParameter returned %f"), parameter );
	log( buf );
}

//-----------------------------------------------------------------------------------------
const String PCLogger::getVstOpCodeName( vstOpCodeGroup group, long opCode )
{
	String retName;

	switch ( group )
	{
		case vstFunctions:
			retName = vstFunctionNames[opCode];
			break;
		case vstDispatcherOpCodes:
			retName = effOpCodes[opCode].opCodeName;
			break;
		case vstAudioMasterOpCodes:
			retName = audioMasterOpCodes[opCode].opCodeName;
			break;
		default:
			retName = "unknown";
			break;
	}

	return retName;
}

//-----------------------------------------------------------------------------------------
int PCLogger::getVstNumOpCodes( vstOpCodeGroup group )
{
	if ( group < 0 || group >= PCLogger::vstNumOpCodeGroups )
		return 0;

	return OpCodeGroupsInfo[group].numOpCodes;
}

//-----------------------------------------------------------------------------------------
long PCLogger::getVstOpCodeCount( vstOpCodeGroup group, long opCode )
{
	return getLogInfo( group, opCode)->count;
}

//-----------------------------------------------------------------------------------------
bool PCLogger::isVstOpCodeEnabled( vstOpCodeGroup group, long opCode )
{
	return getLogInfo( group, opCode)->enabled;
}

//-----------------------------------------------------------------------------------------
void PCLogger::setVstOpCodeEnabled( bool enabled, vstOpCodeGroup group, long opCode )
{
	getLogInfo( group, opCode)->enabled = enabled;
}

//-----------------------------------------------------------------------------------------
void PCLogger::setVstOpCodesEnabled( bool enabled )
{
	for ( int group = 0; group < vstNumOpCodeGroups; group++ )
		for ( long opCode = 0; opCode < OpCodeGroupsInfo[group].numOpCodes; opCode++ )
			OpCodeGroupsInfo[group].debugInfo[opCode].enabled = enabled;
}

//-----------------------------------------------------------------------------------------
void PCLogger::toggleVstOpCodesEnabled()
{
	for ( int group = 0; group < vstNumOpCodeGroups; group++ )
		for ( long opCode = 0; opCode < OpCodeGroupsInfo[group].numOpCodes; opCode++ )
			OpCodeGroupsInfo[group].debugInfo[opCode].enabled = !OpCodeGroupsInfo[group].debugInfo[opCode].enabled;
}

//-----------------------------------------------------------------------------------------
void PCLogger::resetVstOpCodeCounters()
{
	initLogCounters( vstFunctionsLogInfo, vstNumFunctions );
	initLogCounters( vstDispatcherLogInfo, numEffOpCodes );
	initLogCounters( vstAudioMasterLogInfo, numAudioMasterOpCodes );
}

//-----------------------------------------------------------------------------------------
void PCLogger::loadState( const XmlElement& state )
{
	// load opcode preferences
	for ( int group = 0; group < vstNumOpCodeGroups; group++ )
	{
		// search for group element
		XmlElement* groupElement = state.getChildByName( OpCodeGroupsInfo[group].name );

		if ( groupElement != NULL )
		{
			// iterate opCode elements, identify with index and get opCode state
			XmlElement* opCodeElement = groupElement->getFirstChildElement();

			while ( opCodeElement != NULL )
			{
				if ( opCodeElement->hasTagName ( T("opCode") ) )
				{
					long opCode = opCodeElement->getIntAttribute( T("index"), -1 );

					if ( opCode >= 0 && opCode < OpCodeGroupsInfo[group].numOpCodes )
					{
						int enabled = opCodeElement->getIntAttribute( T("enabled"), -1 );
						if ( enabled >= 0 )
							setVstOpCodeEnabled( enabled ? true : false, (PCLogger::vstOpCodeGroup)group, opCode );
					}
				}
				opCodeElement = opCodeElement->getNextElement();
            }

        }
	}

	// load global preferences
	XmlElement* paramsElement = state.getChildByName( T("Parameters") );

	if ( paramsElement )
	{
		int logParams = paramsElement->getIntAttribute( T("logParameters"), -1 );
		if ( logParams >= 0 )
			setLogParameters( logParams ? true : false );

		int logReturnParams = paramsElement->getIntAttribute( T("logReturnParameters"), -1 );
		if ( logReturnParams >= 0 )
			setLogReturnParameters( logReturnParams ? true : false );

		int logTime = paramsElement->getIntAttribute( T("logTime"), -1 );
		if ( logTime >= 0 )
			setLogTime( logTime ? true : false );
	}
}

//-----------------------------------------------------------------------------------------
void PCLogger::saveState( XmlElement& state )
{
	// clean everything
	state.deleteAllChildElements();

	// save opcode preferences
	for ( int group = 0; group < vstNumOpCodeGroups; group++ )
	{
		// create child element for each group
		XmlElement* groupElement = new XmlElement( OpCodeGroupsInfo[group].name );
		state.addChildElement( groupElement );

		// write OpCodes index, name and state
		for ( int opCode = 0; opCode < OpCodeGroupsInfo[group].numOpCodes; opCode++ )
		{
			XmlElement* opCodeElement = new XmlElement( "opCode" );
			opCodeElement->setAttribute( T("index"), opCode );
			opCodeElement->setAttribute( T("name"), getVstOpCodeName( (PCLogger::vstOpCodeGroup)group, opCode ) );
			opCodeElement->setAttribute( T("enabled"), isVstOpCodeEnabled( (PCLogger::vstOpCodeGroup)group, opCode ) );
			groupElement->addChildElement( opCodeElement );
		}
	}

	// save global preferences
	XmlElement* paramsElement = new XmlElement( T("Parameters") );
	paramsElement->setAttribute( T("logParameters"), isLogParameters() );
	paramsElement->setAttribute( T("logReturnParameters"), isLogReturnParameters() );
	paramsElement->setAttribute( T("logTime"), isLogTime() );
	state.addChildElement( paramsElement );
}

